# -*- encoding: utf-8 -*-
# $Id: pstruct.py,v 1.2 2008/01/11 03:21:38 jwp Exp $
##
# copyright 2007, pg/python project.
# http://python.projects.postgresql.org
##
'Postgres data type protocol--packing and unpacking'
import math
import struct

null_sequence = '\xff\xff\xff\xff'

if hasattr(struct, 'Struct'):
	def mk_pack(x):
		s = struct.Struct('!' + x)
		if len(x) > 1:
			return lambda y: s.pack(*y), s.unpack
		else:
			return s.pack, lambda y: s.unpack(y)[0]
else:
	def mk_pack(x):
		l = len(x)
		x = '!' + x
		if l > 1:
			def pack(y):
				return struct.pack(x,*y)
			def unpack(y):
				return struct.unpack(x,y)
		else:
			def pack(y):
				return struct.pack(x,y)
			def unpack(y):
				return struct.unpack(x,y)[0]
		return pack, unpack

def mktimetuple(ts):
	'make a pair of (seconds, microseconds) out of the given double'
	seconds = math.floor(ts)
	return (int(seconds), int(1000000 * (ts - seconds)))

def mktimetuple64(ts):
	'make a pair of (seconds, microseconds) out of the given long'
	seconds = ts // 1000000
	return (seconds, ts - (seconds * 1000000))

def mktime((seconds, ms)):
	'make a double out of the pair of (seconds, microseconds)'
	return float(seconds) + (ms / 1000000.0)

def mktime64((seconds, ms)):
	'make an integer out of the pair of (seconds, microseconds)'
	return seconds * 1000000 + ms

data_to_bool = {'\x01' : True, '\x00' : False}
bool_to_data = {True : '\x01', False : '\x00'}

bool_pack = bool_to_data.__getitem__
bool_unpack = data_to_bool.__getitem__

bitdata_to_bool = {'\x00\x00\x00\x01\x01' : True, '\x00\x00\x00\x01\x00' : False}
bool_to_bitdata = {True : '\x00\x00\x00\x01\x01', False : '\x00\x00\x00\x01\x00'}

bit_pack = bool_to_bitdata.__getitem__
bit_unpack = bitdata_to_bool.__getitem__


longlong_pack, longlong_unpack = mk_pack("q")
long_pack, long_unpack = mk_pack("l")
ulong_pack, ulong_unpack = mk_pack("L")
short_pack, short_unpack = mk_pack("h")
ushort_pack, ushort_unpack = mk_pack("H")
double_pack, double_unpack = mk_pack("d")
float_pack, float_unpack = mk_pack("f")
dd_pack, dd_unpack = mk_pack("dd")
ddd_pack, ddd_unpack = mk_pack("ddd")
dddd_pack, dddd_unpack = mk_pack("dddd")
LH_pack, LH_unpack = mk_pack("LH")
llL_pack, llL_unpack = mk_pack("llL")
qll_pack, qll_unpack = mk_pack("qll")
dll_pack, dll_unpack = mk_pack("dll")
BBBB_pack, BBBB_unpack = mk_pack("BBBB")

dl_pack, dl_unpack = mk_pack("dl")
ql_pack, ql_unpack = mk_pack("ql")


int2_pack, int2_unpack = short_pack, short_unpack
int4_pack, int4_unpack = long_pack, long_unpack
int8_pack, int8_unpack = longlong_pack, longlong_unpack

oid_pack = cid_pack = xid_pack = ulong_pack
oid_unpack = cid_unpack = xid_unpack = ulong_unpack

tid_pack, tid_unpack = LH_pack, LH_unpack

# geometry types
point_pack, point_unpack = dd_pack, dd_unpack
circle_pack, circle_unpack = ddd_pack, ddd_unpack
lseg_pack = box_pack = dddd_pack
lseg_unpack = box_unpack = dddd_unpack

def path_pack(data):
	return struct.pack("!l%dd" %(len(data),), len(data), *data)

def path_unpack(data):
	npoints = long_unpack(data[:4])
	points = struct.unpack("!4x%dd" %(npoints,), data)
	return points
polygon_pack, polygon_unpack = path_pack, path_unpack


# time types
date_pack, date_unpack = long_pack, long_unpack

def time_pack(timetup):
	return double_pack(mktime(timetup))

def time_unpack(data):
	return mktimetuple(double_unpack(data))

def interval_pack((month, day, timetup)):
	return dll_pack((mktime(timetup), day, month))

def interval_unpack(data):
	tim, day, month = dll_unpack(data)
	return (month, day, mktimetuple(tim))

def time64_pack(timetup):
	return longlong_pack(mktime64(timetup))

def time64_unpack(data):
	return mktimetuple64(longlong_unpack(data))

def interval64_pack((month, day, timetup)):
	return qll_pack((mktime64(timetup), day, month))

def interval64_unpack(data):
	tim, day, month = qll_unpack(data)
	return (month, day, mktimetuple64(tim))


def timetz_pack((timetup, tz_offset)):
	return dl_pack((mktime(timetup), tz_offset))

def timetz_unpack(data):
	ts, tz = dl_unpack(data)
	return (mktimetuple(ts), tz)

def timetz64_pack((timetup, tz_offset)):
	return ql_pack((mktime64(timetup), tz_offset))

def timetz64_unpack(data):
	ts, tz = ql_unpack(data)
	return (mktimetuple64(ts), tz)

# oidvectors are 128 bytes, so pack the number of Oids in self
# and justify that to 128 by padding with \x00.
def oidvector_pack(seq):
	return struct.pack("!%dL"%(len(seq),), *seq).ljust(128, '\x00')

def oidvector_unpack(data):
	return struct.unpack("!32L", data)

def int2vector_pack(seq):
	return struct.pack("!%dh"%(len(seq),), *seq).ljust(64, '\x00')

def int2vector_unpack(data):
	return struct.unpack("!32h", data)

def varbit_pack((bits, data)):
	return long_pack(bits) + data

def varbit_unpack(data):
	return long_unpack(data[0:4]), data[4:]

def cidr_pack((family, mask, data)):
	return '%s%s\x01%s%s' %(chr(family), chr(mask), chr(len(data)), data)

def cidr_unpack(data):
	family, mask, is_cidr, size = BBBB_unpack(data[:4])

	rd = data[4:]
	if len(rd) != size:
		raise ValueError("invalid size parameter")

	return (family, mask, rd)

def inet_pack(data):
	if len(data) == 16:
		# 3 = AF_INET6 (family), 128-bit mask
		family_mask = '\x03\x80' # chr(3) + chr(128)
	else:
		# 2 = AF_INET (family), 32-bit mask
		family_mask = '\x02\x20' # chr(2) + chr(32)
	# XXX: is_cidr filled in here (\x00)
	return '%s\x00%s%s' %(family_mask, chr(len(data)), data)

def inet_unpack(data):
	return cidr_unpack(data)[2]

def record_unpack(data):
	"""
	Given serialized record data, return a tuple of tuples of type Oids and
	attributes.
	"""
	columns = long_unpack(data[0:4])
	offset = 4

	for x in xrange(columns):
		typid = oid_unpack(data[offset:offset+4])
		offset += 4

		if data[offset:offset+4] == null_sequence:
			att = None
			offset += 4
		else:
			size = long_unpack(data[offset:offset+4])
			offset += 4
			att = data[offset:offset + size]
			if size < -1 or len(att) != size:
				# XXX: raise 22P03
				raise ValueError("insufficient data left in message")
			offset += size
		yield (typid, att)

	if len(data) - offset != 0:
		raise ValueError("extra data, %d octets, at end of record" %(len(data),))

def record_pack(seq):
	'pack a record given an iterable of (type_oid, data) pairs'
	return long_pack(len(seq)) + ''.join([
		# typid + (null_seq or data)
		oid_pack(x) + (y is None and null_sequence or (long_pack(len(y)) + y))
		for x, y in seq
	])

def elements_pack(elements):
	for x in elements:
		if x is None:
			yield null_sequence
		else:
			yield long_pack(len(x))
			yield x

def array_pack((flags, typid, dlb, elements)):
	'(0, typ oid, (upper_bound, lower_bound, ...), iterable of element_data)'
	header = llL_pack((len(dlb) // 2, flags, typid))
	return header + \
		struct.pack("!%dl" %(len(dlb),), *dlb) + \
		''.join(elements_pack(elements))

def elements_unpack(data, offset):
	data_len = len(data)
	while offset < data_len:
		lend = data[offset:offset+4]
		offset += 4
		if lend == null_sequence:
			yield None
		else:
			sizeof_el = long_unpack(lend)
			yield data[offset:offset+sizeof_el]
			offset += sizeof_el

def array_unpack(data):
	ndim, flags, typid = llL_unpack(data[0:12])
	if ndim < 0:
		raise ValueError("invalid number of dimensions: %d" %(ndim,))
	# "ndim" number of pairs of longs
	end = 4 * 2 * ndim + 12
	# Dimension Bounds
	dlb = struct.unpack("!%dl"%(2 * ndim,), data[12:end])
	return (flags, typid, dlb, elements_unpack(data, end))
